require 'utils'

local edge_colors = {
	['return'] = 'green',
	call = 'red',
	cond = 'blue',
	uncond = 'black',
	subroutine_call = 'pink',
	next = 'black',
	init = 'black',
}

local function nodeprint(x)
	printf("%s [shape=record, label=\"{%07x|instr %d}\", color=%s];\n", x.name, x.addr, x.ninstrs, x.head or 'black');
	for out, edge in pairs(x.outedge) do
		local label
		if edge.type == 'subroutine_call' then
			label = string.format('call %s (%d)', edge.name, edge.n)
		else
			label = string.format('(%d)', edge.n)
		end
			
		printf('%s -> %s [label="%s", color=%s];\n', x.name, out.name, label, edge_colors[edge.type] or edge.type)
	end
end


local nodes = {}
local function connect(a, b, n, type, name)
	local oldedge_a = a.outedge[b]
	local oldedge_b = b.inedge[a]
	if oldedge_a then
		assert(oldedge_a == oldedge_b)
		assert(oldedge_a.type == type)
		assert(oldedge_a.name == name)
		oldedge_a.n = oldedge_a.n + 1
	else
		local edge = { n=n, type=type, name=name }
		a.outedge[b] = edge
		b.inedge[a] = edge
	end
end
local function nedges(x)
	return #table.keys(x)
end
local function disconnect(a, b)
	a.outedge[b] = nil
	b.inedge[a] = nil
end
local function makenode(addr)
	return {
		name = string.format('n%07x', addr),
		addr = addr,
		outedge = {},
		inedge = {},
		ninstrs = 1,
	}
end

local function collapse(a)
	if nedges(a.outedge) ~= 1 then return end
	local b, edge = next(a.outedge)
	if nedges(b.inedge) ~= 1 then return end
	if edge.type ~= 'next' then return end
	-- merge a and b
	disconnect(a, b)
	a.ninstrs = a.ninstrs + b.ninstrs
	for c, edge in pairs(b.outedge) do
		connect(a, c, edge.n, edge.type)
	end
	for c, num in pairs(a.outedge) do
		disconnect(b, c)
	end
	b.deleted = true
	return true
end

local function guess_branch_type(s)
	local first = s:match('^(%w+)')
	if first then
		if first == 'BX' then return 'return' end
		if first == 'B' then return 'uncond' end
		if first == 'BL' or first == 'BLX' then return 'call' end
		if first:match('^B') then return 'cond' end
	end
	return 'next'
end


local function remove_subroutine(addr)
	local subroutine = assert(nodes[addr])
	local callers = {}
	local rp = {}
	local visited = {}
	local rp_tot = 0
	local function go(a)
		if visited[a] then
			return
		end
		visited[a] = true
		for b, edge in pairs(a.outedge) do
			if rp[b] then
				printf('# disconnecting %x %x\n', a.addr, b.addr)
				rp[b] = rp[b] - edge.n
				rp_tot = rp_tot - edge.n
				disconnect(a, b)
				a.head = a.head or 'green'
			end
		end
		for b, edge in pairs(a.outedge) do
			go(b)
		end
	end
	-- find callers
	for caller, edge in pairs(subroutine.inedge) do
		-- grab all edges that are marked with 'call'	
		if edge.type == 'call' then
			callers[caller] = edge
			local return_to = nodes[caller.addr + caller.ninstrs*4]
			if not return_to then
				error('missing return point')	
			end
			assert(not rp[return_to])
			printf('# remove jump to %x with %d calls\n', return_to.addr, edge.n)
			rp[return_to] = edge.n
			rp_tot = rp_tot + edge.n
		end
	end
	-- disconnect callers
	for caller, edge in pairs(callers) do
		disconnect(caller, subroutine)
	end
	-- do a depth-first search
	go(subroutine, false)
	subroutine.head = 'red'
	-- success?
	if rp_tot ~= 0 then
		error('fail to disconnect all')	
	end
	-- reconnect callers
	for caller, edge in pairs(callers) do
		connect(caller, nodes[caller.addr+caller.ninstrs*4], edge.n, 'subroutine_call', subroutine.name)
	end
end

local function analyze_funcalls()
	for _, a in pairs(nodes) do
		for b, edge in pairs(a.outedge) do
			if edge.type == 'call' then
				print('--- CALL ---')
				nodeprint(a)
				local c = nodes[a.addr+4*a.ninstrs]
				if c then
					nodeprint(c)
					local count = 0
					for x, xe in pairs(c.inedge) do
						count = count + xe.n
					end
					print('outcount', edge.n, 'incount', count)
				else
					print('NORETURN!')
				end
			end
		end
	end
end



local prev = 0
local prevtype = 'init'
nodes[prev] = makenode(prev)
nodes[prev].head = 'brown'
for line in io.lines() do
	local addr, instr = line:match('^ARM9 (%x+)%s+%x+%s+(.*)')
	if addr then
		addr = tonumber(addr, 16)
		local node = nodes[addr]
		if not node then
			node = makenode(addr)
			nodes[addr] = node
		end
		local type = guess_branch_type(instr)
		--print('instr', type, instr)
		if prevtype == 'next' and prev + 4 ~= addr then
			prevtype = 'uncond'
		end
		connect(nodes[prev], nodes[addr], 1, prevtype)
		prevtype = type
		prev = addr
	end
end

repeat
	local n_changes = 0
	for _, node in pairs(nodes) do
		if collapse(node) then
			n_changes = n_changes + 1
		end
	end
until n_changes == 0


--analyze_funcalls()
remove_subroutine(0x20196a4)
remove_subroutine(0x200e9dc)
remove_subroutine(0x2019690)
remove_subroutine(0x200ea1c)
remove_subroutine(0x2018054)
remove_subroutine(0x2017b84)
remove_subroutine(0x20180c0)
remove_subroutine(0x2017790)
remove_subroutine(0x2018e6c)
remove_subroutine(0x2074720)

--os.exit(0)
print('digraph G {')
for _, node in pairs(nodes) do
	if not node.deleted then
		nodeprint(node)
	end
end
print('}')
